import requests
from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

from django.db import Error
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from buybackprogram.models import ItemPrices, Owner

from .app_settings import BUYBACKPROGRAM_TASKS_TIME_LIMIT

logger = get_extension_logger(__name__)

NORMAL_TASK_PRIORITY = 4

# Create your tasks here
TASK_DEFAULT_KWARGS = {
    "time_limit": BUYBACKPROGRAM_TASKS_TIME_LIMIT,
}

TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "bind": True,
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": 30,
    },
}


@shared_task
def update_all_prices():

    i = 0
    type_ids = []
    market_data = {}

    # Get all type ids
    prices = ItemPrices.objects.all()

    logger.debug("Price setup starting for %s items from Fuzzworks API" % len(prices))

    # Build suitable bulks to fetch prices from API
    for item in prices:
        type_ids.append(item.eve_type_id)

        i += 1

        if i == 1000:

            response_fuzzwork = requests.get(
                "https://market.fuzzwork.co.uk/aggregates/",
                params={
                    "types": ",".join([str(x) for x in type_ids]),
                    "station": 60003760,
                },
            )

            items_fuzzwork = response_fuzzwork.json()
            market_data.update(items_fuzzwork)

            i = 0
            type_ids.clear()

    # Get leftover data from the bulk
    response_fuzzwork = requests.get(
        "https://market.fuzzwork.co.uk/aggregates/",
        params={
            "types": ",".join([str(x) for x in type_ids]),
            "station": 60003760,
        },
    )

    items_fuzzwork = response_fuzzwork.json()
    market_data.update(items_fuzzwork)

    logger.debug("Market data fetched, starting database update...")
    for price in prices:

        buy = int(float(market_data[str(price.eve_type_id)]["buy"]["max"]))
        sell = int(float(market_data[str(price.eve_type_id)]["sell"]["min"]))

        price.buy = buy
        price.sell = sell
        price.updated = timezone.now()

    try:
        ItemPrices.objects.bulk_update(prices, ["buy", "sell", "updated"])
        logger.debug("All prices succesfully updated")
    except Error as e:
        logger.error("Error updating prices: %s" % e)


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_contracts_for_owner(self, owner_pk):
    """fetches all contracts for owner from ESI"""

    return _get_owner(owner_pk).update_contracts_esi()


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_contracts():
    logger.debug("Starting all contract updates")
    for owner in Owner.objects.all():
        update_contracts_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk},
            priority=NORMAL_TASK_PRIORITY,
        )


def _get_owner(owner_pk: int) -> Owner:
    """returns the owner or raises exception"""
    try:
        owner = Owner.objects.get(pk=owner_pk)
    except Owner.DoesNotExist:
        raise Owner.DoesNotExist(
            "Requested owner with pk {} does not exist".format(owner_pk)
        )
    return owner



from typing import Tuple

from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership, State

# Create your models here.
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.app_settings import BUYBACKPROGRAM_TRACKING_PREFILL

from .decorators import fetch_token_for_owner
from .providers import esi

logger = get_extension_logger(__name__)


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and see own statics."),
            (
                "manage_programs",
                "Can manage own buyback programs and see own program statics.",
            ),
            ("see_all_statics", "Can see all program statics."),
        )


class Owner(models.Model):
    """A corporation that has buyback programs"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_ESI_UNAVAILABLE = 5

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
    ]

    corporation = models.OneToOneField(
        EveCorporationInfo,
        primary_key=True,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        help_text="Character used for retrieving info",
        on_delete=models.deletion.PROTECT,
        related_name="+",
    )

    user = models.ForeignKey(
        User,
        help_text="User that manages the program",
        on_delete=models.deletion.PROTECT,
        related_name="+",
    )

    class Meta:
        default_permissions = ()

    @fetch_token_for_owner(
        [
            "esi-contracts.read_character_contracts.v1",
            "esi-contracts.read_corporation_contracts.v1",
        ]
    )
    def update_contracts_esi(self, token):

        logger.debug("Fetching contracts for %s" % self.character)

        contracts = self._fetch_contracts()

        for contract in contracts:

            # Only get contracts with the correct prefill ticker
            if BUYBACKPROGRAM_TRACKING_PREFILL in contract["title"]:
                obj, created = Contract.objects.update_or_create(
                    assignee_id=contract["assignee_id"],
                    availability=contract["availability"],
                    contract_id=contract["contract_id"],
                    date_completed=contract["date_completed"],
                    date_expired=contract["date_expired"],
                    date_issued=contract["date_issued"],
                    for_corporation=contract["for_corporation"],
                    issuer_corporation_id=contract["issuer_corporation_id"],
                    issuer_id=contract["issuer_id"],
                    price=contract["price"],
                    status=contract["status"],
                    title=contract["title"],
                    volume=contract["volume"],
                    defaults={"contract_id": contract["contract_id"]},
                )

                # If we created an new contract
                if created:
                    logger.debug(
                        "New contract %s created. Starting item fetch"
                        % contract["contract_id"]
                    )

                    character_id = self.character.character.character_id

                    contract_items = esi.client.Contracts.get_characters_character_id_contracts_contract_id_items(
                        character_id=character_id,
                        contract_id=contract["contract_id"],
                        token=token.valid_access_token(),
                    ).results()

                    for item in contract_items:

                        cont = Contract.objects.get(contract_id=contract["contract_id"])
                        itm = EveType.objects.get(pk=item["type_id"])

                        obj, created = ContractItem.objects.update_or_create(
                            contract=cont,
                            eve_type=itm,
                            quantity=item["quantity"],
                            defaults={"contract": cont},
                        )

                    logger.debug(
                        "Updated items for contract %s" % contract["contract_id"]
                    )
                else:
                    logger.debug("Contract %s updated." % contract["contract_id"])

        logger.debug("Fetching corporation contracts for %s" % self.character)

        contracts = self._fetch_corporation_contracts()

        for contract in contracts:

            # Only get contracts with the correct prefill ticker
            if BUYBACKPROGRAM_TRACKING_PREFILL in contract["title"]:
                obj, created = Contract.objects.update_or_create(
                    assignee_id=contract["assignee_id"],
                    availability=contract["availability"],
                    contract_id=contract["contract_id"],
                    date_completed=contract["date_completed"],
                    date_expired=contract["date_expired"],
                    date_issued=contract["date_issued"],
                    for_corporation=contract["for_corporation"],
                    issuer_corporation_id=contract["issuer_corporation_id"],
                    issuer_id=contract["issuer_id"],
                    price=contract["price"],
                    status=contract["status"],
                    title=contract["title"],
                    volume=contract["volume"],
                    defaults={"contract_id": contract["contract_id"]},
                )

                if created:
                    logger.debug(
                        "New corporation contract %s created. Starting item fetch"
                        % contract["contract_id"]
                    )

                    corporation_id = self.character.character.corporation_id

                    contract_items = esi.client.Contracts.get_corporations_corporation_id_contracts_contract_id_items(
                        corporation_id=corporation_id,
                        contract_id=contract["contract_id"],
                        token=token.valid_access_token(),
                    ).results()

                    for item in contract_items:

                        cont = Contract.objects.get(contract_id=contract["contract_id"])
                        itm = EveType.objects.get(pk=item["type_id"])

                        obj, created = ContractItem.objects.update_or_create(
                            contract=cont,
                            eve_type=itm,
                            quantity=item["quantity"],
                            defaults={"contract": cont},
                        )

                    logger.debug(
                        "Updated items for corporation contract %s"
                        % contract["contract_id"]
                    )
                else:
                    logger.debug(
                        "Corporation contract %s updated." % contract["contract_id"]
                    )

    @fetch_token_for_owner(["esi-contracts.read_character_contracts.v1"])
    def _fetch_contracts(self, token) -> list:

        character_id = self.character.character.character_id

        contracts = esi.client.Contracts.get_characters_character_id_contracts(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()

        return contracts

    @fetch_token_for_owner(["esi-contracts.read_corporation_contracts.v1"])
    def _fetch_corporation_contracts(self, token) -> list:

        corporation_id = self.character.character.corporation_id

        contracts = esi.client.Contracts.get_corporations_corporation_id_contracts(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()

        return contracts

    def token(self, scopes=None) -> Tuple[Token, int]:
        """returns a valid Token for the owner"""
        token = None
        error = None

        # abort if character is not configured
        if self.character is None:
            logger.error("%s: No character configured to sync", self)
            error = self.ERROR_NO_CHARACTER

        # abort if character does not have sufficient permissions
        elif self.corporation and not self.character.user.has_perm(
            "buybackprogram.manage_programs"
        ):
            logger.error(
                "%s: This character does not have sufficient permission to sync contracts",
                self,
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        # abort if character does not have sufficient permissions
        elif not self.character.user.has_perm("buybackprogram.manage_programs"):
            logger.error(
                "%s: This character does not have sufficient permission to sync contracts",
                self,
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        else:
            try:
                # get token
                token = (
                    Token.objects.filter(
                        user=self.character.user,
                        character_id=self.character.character.character_id,
                    )
                    .require_scopes(scopes)
                    .require_valid()
                    .first()
                )
            except TokenInvalidError:
                logger.error("%s: Invalid token for fetching calendars", self)
                error = self.ERROR_TOKEN_INVALID
            except TokenExpiredError:
                logger.error("%s: Token expired for fetching calendars", self)
                error = self.ERROR_TOKEN_EXPIRED
            else:
                if not token:
                    logger.error("%s: No token found with sufficient scopes", self)
                    error = self.ERROR_TOKEN_INVALID

        return token, error

    def __str__(self):
        return (
            self.character.character.character_name
            + " ["
            + self.corporation.corporation_ticker
            + "]"
        )


class Location(models.Model):
    """ Location where the buyback program is operated at """

    name = models.CharField(
        max_length=32, help_text="Structure name where the contracts are accepted at"
    )

    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        help_text="System where the buyback structure is located",
        blank=True,
        default=None,
        null=True,
        on_delete=models.deletion.SET_DEFAULT,
        related_name="+",
    )

    def __str__(self):
        return self.eve_solar_system.name + " | " + self.name


class Program(models.Model):
    """An Eve Online buyback program"""

    owner = models.ForeignKey(
        Owner,
        help_text="Player that the contracts will be created for",
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    is_corporation = models.BooleanField(
        default=False,
        help_text="Tick if contracts should be made to corporation instead of player.",
    )

    location = models.ForeignKey(
        Location,
        help_text="Solarystem and station name for contracts.",
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    tax = models.IntegerField(
        default=0,
        blank=False,
        null=False,
        help_text="Default tax is applied on all items unless an item spesific tax is assigned",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    hauling_fuel_cost = models.IntegerField(
        default=0,
        help_text="ISK per m³ that will be removed from the buy price ie. to cover jump freighet fuel costs. <b>Should not be used with price dencity modifier</b>",
    )

    price_dencity_modifier = models.BooleanField(
        default=False,
        help_text="Should we modify buy prices for items with high volume and low value ie. T1 industrial hulls. <b>Should not be used with hauling fuel cost</b>",
    )

    price_dencity_treshold = models.IntegerField(
        default=0,
        null=True,
        help_text="At what ISK/m3 do we start to apply the low isk dencity tax. Tritanium is 500 ISK/m³ @ 5 ISK per unit price. PLEX is 14,5Trillion ISK/m³ @2.9M per unit price.",
    )

    price_dencity_tax = models.IntegerField(
        default=0,
        null=True,
        help_text="How much tax do we apply on the low isk dencity items.",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    allow_all_items = models.BooleanField(
        default=True,
        help_text="If true all items are accepted to the buyback program. You can set extra taxes or disallow individual items from the program item section. If set to false you need to add each accepted item into the program item section. Blueprints are not included in all items.",
    )

    use_refined_value = models.BooleanField(
        default=False,
        help_text="Take refined value into account when calculating prices for ore, ice and moon goo",
    )

    use_compressed_value = models.BooleanField(
        default=False,
        help_text="Take compressed value into account when calculating prices for ore, ice and moon goo",
    )

    use_raw_ore_value = models.BooleanField(
        default=True,
        help_text="Take raw ore value into account when calculating prices for ore, ice and moon goo",
    )

    allow_unpacked_items = models.BooleanField(
        default=False,
        help_text="Do you want to allow unpacked items in this program such as assembled ship hulls?",
    )

    refining_rate = models.IntegerField(
        default=0,
        null=True,
        help_text="Refining rate to be used if refined value is active",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    restricted_to_group = models.ManyToManyField(
        Group,
        blank=True,
        related_name="buybackprogram_require_groups",
        help_text="The group(s) that will be able to see this buyback program. If none is selected program is open for all.",
    )
    restricted_to_state = models.ManyToManyField(
        State,
        blank=True,
        related_name="buybackprogram_require_states",
        help_text="The state(s) that will be able to see this buyback program. If none is selected program is open for all.",
    )

    class Meta:
        default_permissions = ()

    def clean(self):
        super().clean()
        if (
            self.allow_all_items
            and not self.use_refined_value
            and not self.use_compressed_value
            and not self.use_raw_ore_value
        ):
            raise ValidationError(
                "All items are allowed but not a single pricing method for ores is selected. Please use at least one pricing method for ores if all items is allowed."
            )
        if self.price_dencity_modifier and not self.price_dencity_tax:
            raise ValidationError(
                "Price density is used but value for price density tax is missing"
            )
        if self.price_dencity_modifier and not self.price_dencity_treshold:
            raise ValidationError(
                "Price density is used but value for price density threshold is missing"
            )
        if self.use_refined_value and not self.refining_rate:
            raise ValidationError(
                "Refined value is used for ore pricing method but no refining rate is provided. Provide a refining rate to used with this pricing model."
            )


class ProgramItem(models.Model):
    """Items in the buyback program for a corp"""

    program = models.ForeignKey(
        Program,
        on_delete=models.deletion.CASCADE,
        help_text="What program do these items belong to",
    )
    item_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Select item for special tax",
    )
    item_tax = models.IntegerField(
        default=0,
        null=True,
        help_text="How much tax do we add on top of the base tax for this item. Can also be negative.",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    disallow_item = models.BooleanField(
        default=False,
        help_text="You can disallow an item from a buyback location. It will return 0 price if disallowed.",
    )

    class Meta:
        default_permissions = ()
        unique_together = ["program", "item_type"]

    def save(self, *args, **kwargs):
        if not self.item_tax:
            self.item_tax = self.program.tax
        super(ProgramItem, self).save(*args, **kwargs)


class ItemPrices(models.Model):
    eve_type = models.OneToOneField(
        EveType,
        on_delete=models.deletion.CASCADE,
        unique=True,
    )
    buy = models.BigIntegerField()
    sell = models.BigIntegerField()
    updated = models.DateTimeField()


class Tracking(models.Model):
    program = models.ForeignKey(
        Program,
        null=True,
        on_delete=models.deletion.SET_NULL,
        related_name="+",
    )
    issuer_user = models.ForeignKey(
        User,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    value = models.BigIntegerField(null=False)
    taxes = models.BigIntegerField(null=False)
    hauling_cost = models.BigIntegerField(null=False)
    donation = models.BigIntegerField(null=True, blank=True)
    net_price = models.BigIntegerField(null=False)
    tracking_number = models.CharField(max_length=20)


class TrackingItem(models.Model):

    tracking = models.ForeignKey(
        Tracking,
        on_delete=models.deletion.CASCADE,
        help_text="What tracking do these items belong to",
    )

    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Item type information",
    )

    buy_value = models.BigIntegerField(null=False)

    quantity = models.IntegerField()


class Contract(models.Model):

    assignee_id = models.IntegerField()
    availability = models.CharField(max_length=20)
    contract_id = models.IntegerField()
    date_completed = models.DateTimeField(null=True)
    date_expired = models.DateTimeField(null=True)
    date_issued = models.DateTimeField()
    for_corporation = models.BooleanField()
    issuer_corporation_id = models.IntegerField()
    issuer_id = models.IntegerField()
    price = models.BigIntegerField()
    status = models.CharField(max_length=30)
    title = models.CharField(max_length=128)
    volume = models.BigIntegerField()


class ContractItem(models.Model):

    contract = models.ForeignKey(
        Contract,
        on_delete=models.deletion.CASCADE,
        help_text="What contract do these items belong to",
    )

    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Item type information",
    )

    quantity = models.IntegerField()



# Generated by Django 3.2.8 on 2021-12-21 17:50

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveonline", "0015_factions"),
        ("authentication", "0019_merge_20211026_0919"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app and see own statics."),
                    (
                        "manage_programs",
                        "Can manage own buyback programs and see own program statics.",
                    ),
                    ("see_all_statics", "Can see all program statics."),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Contract",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("assignee_id", models.IntegerField()),
                ("availability", models.CharField(max_length=20)),
                ("contract_id", models.IntegerField()),
                ("date_completed", models.DateTimeField(null=True)),
                ("date_expired", models.DateTimeField(null=True)),
                ("date_issued", models.DateTimeField()),
                ("for_corporation", models.BooleanField()),
                ("issuer_corporation_id", models.IntegerField()),
                ("issuer_id", models.IntegerField()),
                ("price", models.BigIntegerField()),
                ("status", models.CharField(max_length=30)),
                ("title", models.CharField(max_length=128)),
                ("volume", models.BigIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Structure name where the contracts are accepted at",
                        max_length=32,
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="System where the buyback structure is located",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        help_text="Character used for retrieving info",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User that manages the program",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Program",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_corporation",
                    models.BooleanField(
                        default=False,
                        help_text="Tick if contracts should be made to corporation instead of player.",
                    ),
                ),
                (
                    "tax",
                    models.IntegerField(
                        default=0,
                        help_text="Default tax is applied on all items unless an item spesific tax is assigned",
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                    ),
                ),
                (
                    "hauling_fuel_cost",
                    models.IntegerField(
                        default=0,
                        help_text="ISK per m³ that will be removed from the buy price ie. to cover jump freighet fuel costs. <b>Should not be used with price dencity modifier</b>",
                    ),
                ),
                (
                    "price_dencity_modifier",
                    models.BooleanField(
                        default=False,
                        help_text="Should we modify buy prices for items with high volume and low value ie. T1 industrial hulls. <b>Should not be used with hauling fuel cost</b>",
                    ),
                ),
                (
                    "price_dencity_treshold",
                    models.IntegerField(
                        default=0,
                        help_text="At what ISK/m3 do we start to apply the low isk dencity tax. Tritanium is 500 ISK/m³ @ 5 ISK per unit price. PLEX is 14,5Trillion ISK/m³ @2.9M per unit price.",
                        null=True,
                    ),
                ),
                (
                    "price_dencity_tax",
                    models.IntegerField(
                        default=0,
                        help_text="How much tax do we apply on the low isk dencity items.",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                    ),
                ),
                (
                    "allow_all_items",
                    models.BooleanField(
                        default=True,
                        help_text="If true all items are accepted to the buyback program. You can set extra taxes or disallow individual items from the program item section. If set to false you need to add each accepted item into the program item section. Blueprints are not included in all items.",
                    ),
                ),
                (
                    "use_refined_value",
                    models.BooleanField(
                        default=False,
                        help_text="Take refined value into account when calculating prices for ore, ice and moon goo",
                    ),
                ),
                (
                    "use_compressed_value",
                    models.BooleanField(
                        default=False,
                        help_text="Take compressed value into account when calculating prices for ore, ice and moon goo",
                    ),
                ),
                (
                    "use_raw_ore_value",
                    models.BooleanField(
                        default=True,
                        help_text="Take raw ore value into account when calculating prices for ore, ice and moon goo",
                    ),
                ),
                (
                    "allow_unpacked_items",
                    models.BooleanField(
                        default=False,
                        help_text="Do you want to allow unpacked items in this program such as assembled ship hulls?",
                    ),
                ),
                (
                    "refining_rate",
                    models.IntegerField(
                        default=0,
                        help_text="Refining rate to be used if refined value is active",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        help_text="Solarystem and station name for contracts.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="buybackprogram.location",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Player that the contracts will be created for",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="buybackprogram.owner",
                    ),
                ),
                (
                    "restricted_to_group",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The group(s) that will be able to see this buyback program. If none is selected program is open for all.",
                        related_name="buybackprogram_require_groups",
                        to="auth.Group",
                    ),
                ),
                (
                    "restricted_to_state",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The state(s) that will be able to see this buyback program. If none is selected program is open for all.",
                        related_name="buybackprogram_require_states",
                        to="authentication.State",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Tracking",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.BigIntegerField()),
                ("taxes", models.BigIntegerField()),
                ("hauling_cost", models.BigIntegerField()),
                ("donation", models.BigIntegerField(blank=True, null=True)),
                ("net_price", models.BigIntegerField()),
                ("tracking_number", models.CharField(max_length=20)),
                (
                    "issuer_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="buybackprogram.program",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TrackingItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy_value", models.BigIntegerField()),
                ("quantity", models.IntegerField()),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="Item type information",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "tracking",
                    models.ForeignKey(
                        help_text="What tracking do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.tracking",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ItemPrices",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy", models.BigIntegerField()),
                ("sell", models.BigIntegerField()),
                ("updated", models.DateTimeField()),
                (
                    "eve_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ContractItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.IntegerField()),
                (
                    "contract",
                    models.ForeignKey(
                        help_text="What contract do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.contract",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="Item type information",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ProgramItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "item_tax",
                    models.IntegerField(
                        default=0,
                        help_text="How much tax do we add on top of the base tax for this item. Can also be negative.",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                    ),
                ),
                (
                    "disallow_item",
                    models.BooleanField(
                        default=False,
                        help_text="You can disallow an item from a buyback location. It will return 0 price if disallowed.",
                    ),
                ),
                (
                    "item_type",
                    models.ForeignKey(
                        help_text="Select item for special tax",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        help_text="What program do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.program",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
                "unique_together": {("program", "item_type")},
            },
        ),
    ]



from django.contrib.auth.decorators import login_required, permission_required
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy
from esi.decorators import token_required

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.forms import LocationForm, ProgramForm, ProgramItemForm
from buybackprogram.models import Location, Owner, Program, ProgramItem
from buybackprogram.utils import messages_plus

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.manage_programs")
@token_required(
    scopes=[
        "esi-contracts.read_character_contracts.v1",
        "esi-contracts.read_corporation_contracts.v1",
    ]
)
def setup(request, token):
    success = True
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    try:
        owned_char = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy(
                    "You can only use your main or alt characters "
                    "to add corporations. "
                    "However, character %s is neither. "
                )
                % format_html("<strong>{}</strong>", token_char.character_name)
            ),
        )
        success = False

    if success:
        try:
            corporation = EveCorporationInfo.objects.get(
                corporation_id=token_char.corporation_id
            )
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create_corporation(
                token_char.corporation_id
            )

        with transaction.atomic():
            owner, _ = Owner.objects.update_or_create(
                corporation=corporation, character=owned_char, user=request.user
            )

            owner.save()

        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "%(corporation)s has been added with %(character)s "
                    "as program manager. "
                )
                % {
                    "corporation": format_html(
                        "<strong>{}</strong>", owner.corporation
                    ),
                    "character": format_html(
                        "<strong>{}</strong>", owner.character.character.character_name
                    ),
                }
            ),
        )

    return redirect("buybackprogram:index")


@login_required
@permission_required("buybackprogram.manage_programs")
def program_add(request):
    form = ProgramForm(request.POST or None, user=request.user)

    if request.POST and form.is_valid():

        form = ProgramForm(request.POST, user=request.user)

        new_program = form.save()

        messages_plus.success(
            request,
            format_html(
                gettext_lazy("New program created at %(location)s")
                % {
                    "location": format_html(
                        "<strong>{}</strong>", new_program.location
                    ),
                }
            ),
        )

        return HttpResponseRedirect(reverse("buybackprogram:index"))

    return render(request, "buybackprogram/program_add.html", {"form": form})


@login_required
@permission_required("buybackprogram.manage_programs")
def program_edit(request, program_pk):

    instance = Program.objects.get(pk=program_pk)

    if request.method == "POST":
        form = ProgramForm(request.POST, user=request.user)

        logger.debug(
            "Received POST request containing update program form, is valid: %s"
            % form.is_valid()
        )

        if form.is_valid():

            form = ProgramForm(request.POST, instance=instance, user=request.user)

            updated_program = form.save()

            messages_plus.success(
                request,
                format_html(
                    gettext_lazy("Program updated at %(location)s")
                    % {
                        "location": format_html(
                            "<strong>{}</strong>", updated_program.location
                        ),
                    }
                ),
            )

            return HttpResponseRedirect(reverse("buybackprogram:index"))

    else:

        form = ProgramForm(instance=instance, user=request.user)

    context = {
        "program": instance,
        "form": form,
    }

    return render(request, "buybackprogram/program_edit.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def location_add(request):

    locations = Location.objects.all()

    if request.method != "POST":
        form = LocationForm()
    else:
        form = LocationForm(
            request.POST,
            value=int(request.POST["eve_solar_system"]),
        )

        if form.is_valid():
            eve_solar_system = form.cleaned_data["eve_solar_system"]
            name = form.cleaned_data["name"]

            created = Location.objects.update_or_create(
                eve_solar_system=eve_solar_system,
                name=name,
                defaults={
                    "eve_solar_system": eve_solar_system,
                    "name": name,
                },
            )

            if created:
                messages_plus.success(
                    request,
                    format_html(
                        "Added location for <strong>{}</strong> in system <strong>{}</strong>",
                        name,
                        eve_solar_system,
                    ),
                )

            return HttpResponseRedirect(request.path_info)

    context = {
        "locations": locations,
        "form": form,
    }

    return render(request, "buybackprogram/location_add.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def program_edit_item(request, program_pk):
    program = Program.objects.get(pk=program_pk)

    program_items = ProgramItem.objects.filter(program=program)

    if request.method != "POST":
        form = ProgramItemForm()
    else:
        form = ProgramItemForm(
            request.POST,
            value=int(request.POST["item_type"]),
        )

        if form.is_valid():
            item_type = form.cleaned_data["item_type"]
            item_tax = form.cleaned_data["item_tax"]
            disallow_item = form.cleaned_data["disallow_item"]

            created = ProgramItem.objects.update_or_create(
                item_type=item_type,
                program=program,
                defaults={
                    "item_tax": item_tax,
                    "disallow_item": disallow_item,
                },
            )

            if created:
                messages_plus.success(
                    request,
                    format_html(
                        "Added <strong>{}</strong> to program with <strong>{}</strong> % tax",
                        item_type,
                        item_tax,
                    ),
                )

            return HttpResponseRedirect(request.path_info)

    context = {
        "program": program,
        "program_items": program_items,
        "form": form,
    }

    return render(request, "buybackprogram/program_edit_item.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_remove(request, program_pk):

    program = Program.objects.get(pk=program_pk)

    if program.owner.user == request.user:
        program.delete()

        messages_plus.warning(
            request,
            format_html(
                gettext_lazy("Deleted program for %(location)s")
                % {
                    "location": format_html("<strong>{}</strong>", program.location),
                }
            ),
        )

    else:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy(
                    "You do not own this program and thus you can't delete it."
                )
            ),
        )

    return redirect("buybackprogram:index")


@login_required
@permission_required("buybackprogram.manage_programs")
def program_item_remove(request, item_pk, program_pk):

    program_item = ProgramItem.objects.get(item_type=item_pk)

    name = program_item.item_type

    program_item.delete()

    messages_plus.warning(
        request,
        format_html(
            "Deleted <strong>{}</strong> from program",
            name,
        ),
    )

    return redirect("buybackprogram:program_edit_item", program_pk)



from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.shortcuts import redirect, render
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from ..models import Contract, ContractItem, Tracking, TrackingItem

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def my_stats(request):

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    tracking = Tracking.objects.all()

    tracking_numbers = tracking.values_list("tracking_number", flat=True)

    contracts = Contract.objects.filter(
        issuer_id__in=characters,
        title__in=tracking_numbers,
    )

    for contract in contracts:

        if contract.status == "outstanding":
            values["outstanding"] += contract.price
            values["outstanding_count"] += 1
        if contract.status == "finished":
            values["finished"] += contract.price
            values["finished_count"] += 1

        contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

        contract.items = ContractItem.objects.filter(contract=contract)

        contract_tracking = tracking.filter(tracking_number=contract.title).first()

        if contract_tracking.net_price != contract.price:
            note = {
                "icon": "fa-skull-crossbones",
                "color": "red",
                "message": "Tracked price does not match contract price. Either a mistake in the copy pasted values or seller has changed contracted items after calculating the price",
            }

            contract.note = note

    context = {
        "contracts": contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_stats(request):

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    logger.debug("Got characters for manager: %s" % characters)

    corporations = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__corporation_id", flat=True
    )

    logger.debug("Got corporations for manager: %s" % corporations)

    tracking = Tracking.objects.all()

    tracking_numbers = tracking.values_list("tracking_number", flat=True)

    contracts = Contract.objects.filter(
        Q(assignee_id__in=characters) | Q(assignee_id__in=corporations),
        title__in=tracking_numbers,
    )

    logger.debug("Got contracts for manager: %s" % contracts)

    for contract in contracts:

        if contract.status == "outstanding":
            values["outstanding"] += contract.price
            values["outstanding_count"] += 1
        if contract.status == "finished":
            values["finished"] += contract.price
            values["finished_count"] += 1

        contract.notes = []

        try:
            issuer_character = CharacterOwnership.objects.get(
                character__character_id=contract.issuer_id
            )
            logger.debug("Got issuer character from auth: %s" % issuer_character.user)

        except CharacterOwnership.DoesNotExist:
            issuer_character = False
            logger.debug("Contract issuer not registered on AUTH")

            note = {
                "icon": "fa-question",
                "color": "orange",
                "message": "Issuer not registered on AUTH. Possibly an unregistered alt.",
            }

            contract.notes.append(note)

        contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

        contract.assignee_name = EveEntity.objects.resolve_name(contract.assignee_id)

        contract.items = ContractItem.objects.filter(contract=contract)

        contract_tracking = tracking.filter(tracking_number=contract.title).first()

        if contract_tracking.net_price != contract.price:
            note = {
                "icon": "fa-skull-crossbones",
                "color": "red",
                "message": "Tracked price for %s does not match contract price. See details for more information"
                % contract_tracking.tracking_number,
            }

            contract.notes.append(note)

        if (
            contract.assignee_id in corporations
            and not contract_tracking.program.is_corporation
        ):
            note = {
                "icon": "fa-home",
                "color": "orange",
                "message": "Contract %s is made for your corporation while they should be made directly to your character in this program."
                % contract_tracking.tracking_number,
            }

            contract.notes.append(note)

        if (
            contract.assignee_id not in corporations
            and contract_tracking.program.is_corporation
        ):
            note = {
                "icon": "fa-user",
                "color": "orange",
                "message": "Contract %s is made for your character while they should be made to your corporation in this program."
                % contract_tracking.tracking_number,
            }

            contract.notes.append(note)

    context = {
        "contracts": contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/program_stats.html", context)


@login_required
@permission_required("buybackprogram.manage_all_programs")
def program_stats_all(request):

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    tracking = Tracking.objects.all()

    tracking_numbers = tracking.values_list("tracking_number", flat=True)

    contracts = Contract.objects.filter(
        title__in=tracking_numbers,
    )

    for contract in contracts:

        if contract.status == "outstanding":
            values["outstanding"] += contract.price
            values["outstanding_count"] += 1
        if contract.status == "finished":
            values["finished"] += contract.price
            values["finished_count"] += 1

        contract.notes = []

        try:
            issuer_character = CharacterOwnership.objects.get(
                character__character_id=contract.issuer_id
            )
            logger.debug("Got issuer character from auth: %s" % issuer_character.user)

        except CharacterOwnership.DoesNotExist:
            issuer_character = False
            logger.debug("Contract issuer not registered on AUTH")

            note = {
                "icon": "fa-question",
                "color": "orange",
                "message": "Issuer not registered on AUTH. Possibly an unregistered alt.",
            }

            contract.notes.append(note)

        contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

        contract.assignee_name = EveEntity.objects.resolve_name(contract.assignee_id)

        contract.items = ContractItem.objects.filter(contract=contract)

        contract_tracking = tracking.filter(tracking_number=contract.title).first()

        if contract_tracking.net_price != contract.price:
            note = {
                "icon": "fa-skull-crossbones",
                "color": "red",
                "message": "Tracked price for %s does not match contract price. See details for more information"
                % contract_tracking.tracking_number,
            }

            contract.notes.append(note)

    context = {
        "contracts": contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/program_stats_all.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def contract_details(request, contract_title):
    try:

        notes = []

        contract = Contract.objects.get(title=contract_title)

        contract_items = ContractItem.objects.filter(contract=contract)

        tracking = Tracking.objects.get(
            tracking_number=contract_title,
        )

        tracking_items = TrackingItem.objects.filter(tracking=tracking)

        if contract.price != tracking.net_price:
            note = {
                "icon": "fa-skull-crossbones",
                "color": "alert-danger",
                "message": "Tracked price does not match contract price. You have either made an mistake in the tracking number or the contract price copy paste. Please remake contract.",
            }
            notes.append(note)

        context = {
            "notes": notes,
            "contract": contract,
            "contract_items": contract_items,
            "tracking": tracking,
            "tracking_items": tracking_items,
        }

        return render(request, "buybackprogram/contract_details.html", context)

    except Contract.DoesNotExist:
        return redirect("buybackprogram/stats.html")
